@@ -3,7 +3,7 @@
 #
 #  (C) Paul Evans, 2019-2022 -- leonerd@leonerd.org.uk
 
-package Object::Pad 0.61;
+package Object::Pad 0.78;
 
 use v5.14;
 use warnings;
@@ -28,6 +28,8 @@ require mro;
 
 require Object::Pad::MOP::Class;
 
+=encoding UTF-8
+
 =head1 NAME
 
 C<Object::Pad> - a simple syntax for lexical field-based objects
@@ -40,8 +42,8 @@ On perl version 5.26 onwards:
    use Object::Pad;
 
    class Point {
-      has $x :param = 0;
-      has $y :param = 0;
+      field $x :param = 0;
+      field $y :param = 0;
 
       method move ($dX, $dY) {
          $x += $dX;
@@ -60,8 +62,8 @@ Or, for older perls that lack signatures:
    use Object::Pad;
 
    class Point {
-      has $x :param = 0;
-      has $y :param = 0;
+      field $x :param = 0;
+      field $y :param = 0;
 
       method move {
          my ($dX, $dY) = @_;
@@ -81,9 +83,9 @@ Or, for older perls that lack signatures:
 This module provides a simple syntax for creating object classes, which uses
 private variables that look like lexicals as object member fields.
 
-B<WARNING> This module is still experimental. The parts that currently exist
-do seem to work reliably but much of the design is still evolving, and many
-features and have yet to be implemented. I don't yet guarantee I won't have to
+While most of this module has evolved into a stable state in practice, parts
+remain B<experimental> because the design is still evolving, and many features
+and ideas have yet to implemented. I don't yet guarantee I won't have to
 change existing details in order to continue its development. Feel free to try
 it out in experimental or newly-developed code, but don't complain if a later
 version is incompatible with your current code and you'll have to change it.
@@ -93,6 +95,35 @@ The more feedback you provide in terms of what features you are using, what
 you find works, and what doesn't, will help the ongoing development and
 hopefully eventual stability of the design. See the L</FEEDBACK> section.
 
+=head2 Experimental Features
+
+I<Since version 0.63.>
+
+Some of the features of this module are currently marked as experimental. They
+will provoke warnings in the C<experimental> category, unless silenced.
+
+You can silence this with C<no warnings 'experimental'> but then that will
+silence every experimental warning, which may hide others unintentionally. For
+a more fine-grained approach you can instead use the import line for this
+module to only silence the module's warnings selectively:
+
+   use Object::Pad ':experimental(init_expr)';
+
+   use Object::Pad ':experimental(mop)';
+
+   use Object::Pad ':experimental(custom_field_attr)';
+
+   use Object::Pad ':experimental(adjust_params)';
+
+   use Object::Pad ':experimental';  # all of the above
+
+I<Since version 0.64.>
+
+Multiple experimental features can be enabled at once by giving multiple names
+in the parens, separated by spaces:
+
+   use Object::Pad ':experimental(init_expr mop)';
+
 =head2 Automatic Construction
 
 Classes are automatically provided with a constructor method, called C<new>,
@@ -130,20 +161,16 @@ not need to chain to the C<SUPER> class first; this is handled automatically.
 
 =head3 The ADJUST phase
 
-Next, the C<ADJUST> and C<ADJUSTPARAMS> block of every component class is
-invoked. This happens after the fields are assigned their initial values and
-the C<BUILD> blocks have been run.
-
-Note also that both C<ADJUST> and C<ADJUSTPARAMS> blocks happen at the same
-time, in declaration order. The C<ADJUSTPARAMS> blocks do not form their own
-separate phase.
+Next, the C<ADJUST> block of every component class is invoked. This happens
+after the fields are assigned their initial values and the C<BUILD> blocks
+have been run.
 
 =head3 The strict-checking phase
 
 Finally, before the object is returned, if the L</:strict(params)> class
 attribute is present, then the constructor will throw an exception if there
 are any remaining named arguments left over after assigning them to fields as
-per C<:param> declarations, and running any C<ADJUSTPARAMS> blocks.
+per C<:param> declarations, and running any C<ADJUST> blocks.
 
 =head1 KEYWORDS
 
@@ -183,10 +210,10 @@ I<Since version 0.41.>
       ...
    }
 
-Prior to version 0.41 this was called C<extends>, which is currently
-recognised as a compatibility synonym. Both C<extends> and C<isa> keywords are
-now discouraged, in favour of the L</:isa> attribute which is preferred
-because it follows a more standard grammar without this special-case.
+Prior to version 0.41 this was called C<extends>, but I<since version 0.73>
+this is no longer recognised. The C<isa> keyword is now deprecated, in favour
+of the L</:isa> attribute which is preferred because it follows a more
+standard grammar without this special-case.
 
 One or more roles can be composed into the class by the keyword C<does>
 
@@ -196,10 +223,10 @@ I<Since version 0.41.>
       ...
    }
 
-Prior to version 0.41 this was called C<implements>, which is currently
-recognised as a compatibility synonym. Both C<implements> and C<does> keywords
-are now discouraged, in favour of the L</:does> attribute which is preferred
-because it follows a more standard grammar without this special-case.
+Prior to version 0.41 this was called C<implements>, but I<since version 0.73>
+this is no longer recognised. The C<does> keyword is now deprecated, in favour
+of the L</:does> attribute which is preferred because it follows a more
+standard grammar without this special-case.
 
 An optional list of attributes may be supplied in similar syntax as for subs
 or lexical variables. (These are annotations about the class itself; the
@@ -229,7 +256,7 @@ supported.
 If the package providing the superclass does not exist, an attempt is made to
 load it by code equivalent to
 
-   require CLASS ();
+   require CLASS;
 
 and thus it must either already exist, or be locatable via the usual C<@INC>
 mechanisms.
@@ -268,7 +295,7 @@ following values:
    :repr(native)
 
 The native representation. This is an opaque representation type whose
-contents are not specified. It only works for classes whose entire inheritence
+contents are not specified. It only works for classes whose entire inheritance
 hierarchy is built only from classes based on C<Object::Pad>.
 
    :repr(HASH)
@@ -316,7 +343,7 @@ I<Since version 0.43.>
 Can only be applied to classes that contain no C<BUILD> blocks. If set, then
 the constructor will complain about any unrecognised named arguments passed to
 it (i.e. names that do not correspond to the C<:param> of any defined field
-and left unconsumed by any C<ADJUSTPARAMS> block).
+and left unconsumed by any C<ADJUST> block).
 
 Since C<BUILD> blocks can inspect the arguments arbitrarily, the presence of
 any such block means the constructor cannot determine which named arguments
@@ -338,7 +365,7 @@ C<BUILD> blocks are present.
 I<Since version 0.32.>
 
 Similar to C<class>, but provides a package that defines a new role. A role
-acts simliar to a class in some respects, and differently in others.
+acts similar to a class in some respects, and differently in others.
 
 Like a class, a role can have a version, and named methods.
 
@@ -357,17 +384,17 @@ implements the role.
       requires METHOD;
    }
 
-A role can provide instance fields. These are visible to any C<BUILD> blocks
+A role can provide instance fields. These are visible to any C<ADJUST> blocks
 or methods provided by that role.
 
 I<Since version 0.33.>
 
    role Name {
-      has $field;
+      field $f;
 
-      BUILD { $field = "a value" }
+      ADJUST { $f = "a value"; }
 
-      method field { return $field }
+      method field { return $f; }
    }
 
 I<Since version 0.57> a role can declare that it provides another role:
@@ -401,7 +428,7 @@ written in classical Perl that has not yet been rewritten to use
 C<Object::Pad>.
 
 The tradeoff is that a C<:compat(invokable)> role may not create field data
-using the L</has> keyword. Whatever behaviours the role wishes to perform
+using the L</field> keyword. Whatever behaviours the role wishes to perform
 must be provided only by calling other methods on C<$self>, or perhaps by
 making assumptions about the representation type of instances.
 
@@ -410,17 +437,22 @@ upgrade of existing classical Perl code into using C<Object::Pad>. When all
 existing code is using C<Object::Pad> then this attribute can be removed from
 the role.
 
-=head2 has
+=head2 field
 
-   has $var;
-   has @var;
-   has %var;
+   field $var;
+   field @var;
+   field %var;
 
-   has $var :ATTR ATTR...;
+   field $var :ATTR ATTR...;
 
-   has $var = EXPR;
+   field $var = EXPR;
 
-   has $var { BLOCK };
+   field $var //= EXPR;
+   field $var ||= EXPR;
+
+   field $var { BLOCK }
+
+I<Since version 0.66.>
 
 Declares that the instances of the class or role have a member field of the
 given name. This member field will be accessible as a lexical variable within
@@ -435,35 +467,6 @@ to. In order to provide access to them a class may wish to use L</method> to
 create an accessor, or use the attributes such as L</:reader> to get one
 generated.
 
-A scalar field may provide a expression that gives an initialisation value,
-which will be assigned into the field of every instance during the constructor
-before the C<BUILD> blocks are invoked. I<Since version 0.29> this expression
-does not have to be a compiletime constant, though it is evaluated exactly
-once, at runtime, after the class definition has been parsed. It is not
-evaluated individually for every object instance of that class. I<Since
-version 0.54> this is also permitted on array and hash fields.
-
-=head3 Field Initialiser Blocks
-
-I<Since version 0.54> a deferred statement block is also permitted, on any
-field variable type. This is an B<experimental> feature that permits code to
-be executed as part of the instance constructor, rather than running just
-once when the class is set up. Code in a field initialisation block is
-roughly equivalent to being placed in a C<BUILD> or C<ADJUST> block.
-
-Control flow that attempts to leave a field initialiser block is not
-permitted. This includes any C<return> expression, any C<next/last/redo>
-outside of a loop, with a dynamically-calculated label expression, or with a
-label that it doesn't appear in. C<goto> statements are also currently
-forbidden, though known-safe ones may be permitted in future.
-
-Loop control expressions that are known at compiletime to affect a loop that
-they appear within are permitted.
-
-   has $field { foreach(@list) { next; } }       # this is fine
-
-   has $field { LOOP: while(1) { last LOOP; } }  # this is fine too
-
 The following field attributes are supported:
 
 =head3 :reader, :reader(NAME)
@@ -474,10 +477,10 @@ Generates a reader method to return the current value of the field. If no name
 is given, the name of the field is used. A single prefix character C<_> will
 be removed if present.
 
-   has $field :reader;
+   field $x :reader;
 
    # equivalent to
-   has $field;  method field { return $field }
+   field $x;  method x { return $x }
 
 I<Since version 0.55> these are permitted on any field type, but prior
 versions only allowed them on scalar fields. The reader method behaves
@@ -485,7 +488,7 @@ identically to how a lexical variable would behave in the same context; namely
 returning a list of values from an array or key/value pairs from a hash when
 in list context, or the number of items or keys when in scalar context.
 
-   has @items :reader;
+   field @items :reader;
 
    foreach my $item ( $obj->items ) { ... }   # iterates the list of items
 
@@ -499,10 +502,10 @@ Generates a writer method to set a new value of the field from its arguments.
 If no name is given, the name of the field is used prefixed by C<set_>. A
 single prefix character C<_> will be removed if present.
 
-   has $field :writer;
+   field $x :writer;
 
    # equivalent to
-   has $field;  method set_field { $field = shift; return $self }
+   field $x;  method set_x { $x = shift; return $self }
 
 I<Since version 0.28> a generated writer method will return the object
 invocant itself, allowing a chaining style.
@@ -524,10 +527,10 @@ Generates an lvalue mutator method to return or set the value of the field.
 These are only permitted for scalar fields. If no name is given, the name of
 the field is used. A single prefix character C<_> will be removed if present.
 
-   has $field :mutator;
+   field $x :mutator;
 
    # equivalent to
-   has $field;  method field :lvalue { $field }
+   field $x;  method x :lvalue { $x }
 
 I<Since version 0.28> all of these generated accessor methods will include
 argument checking similar to that used by subroutine signatures, to ensure the
@@ -549,14 +552,14 @@ C<undef>). If no argument is passed (i.e. C<scalar @_> is false) then the
 field is not modified. In either case, the value of the field is then
 returned.
 
-   has $field :accessor;
+   field $x :accessor;
 
    # equivalent to
-   has $field;
+   field $x;
 
    method field {
-      $field = shift if @_;
-      return $field;
+      $x = shift if @_;
+      return $x;
    }
 
 =head3 :weak
@@ -568,7 +571,8 @@ contains a reference. This applies to within the constructor if C<:param> is
 given, and to a C<:writer> accessor method. Note that this I<only> applies to
 automatically generated code; not normal code written in regular method
 bodies. If you assign into the field variable you must remember to call
-C<Scalar::Util::weaken> yourself.
+C<Scalar::Util::weaken> (or C<builtin::weaken> on Perl 5.36 or above)
+yourself.
 
 =head3 :param, :param(NAME)
 
@@ -584,8 +588,8 @@ to invoke the constructor without a named argument for this will throw an
 exception. In order to make a parameter optional, make sure to give it a
 default expression - even if that expression is C<undef>:
 
-   has $x :param;          # this is required
-   has $z :param = undef;  # this is optional
+   field $x :param;          # this is required
+   field $z :param = undef;  # this is optional
 
 Any field that has a C<:param> and an initialisation block will only run the
 code in the block if required by the constructor. If a named parameter is
@@ -595,6 +599,91 @@ executed.
 Values for fields are assigned by the constructor before any C<BUILD> blocks
 are invoked.
 
+=head3 Field Initialiser Expressions
+
+I<Since version 0.54> a deferred statement block is also permitted, on any
+field variable type. This permits code to be executed as part of the instance
+constructor, rather than running just once when the class is set up. Code in a
+field initialisation block is roughly equivalent to being placed in a C<BUILD>
+or C<ADJUST> block.
+
+I<Since version 0.73> this may also be written as a plain expression
+introduced by an equals symbol (C<=>). This is equivalent to using a block.
+Note carefully: the equals symbol is part of the C<field> syntax; it is I<not>
+simply a runtime assignment operator that happens once at the time the class
+is declared. Just like the block form describe above, the expression is
+evaluated during the constructor of every instance.
+
+I<Since version 0.74> this expression may also be written using a defined-or
+or logical-or assignment operator (C<//=> or C<||=>). In these case, the
+default expression will be evaluated and assigned if the caller did not pass a
+value to the constructor at all, or if the value passed was undef (for C<//=>)
+or false (for C<||=>). For most scalar parameters, where C<undef> is not a
+valid value, you probably wanted to use C<//=> to assign defaults.
+
+   class Action {
+      field $timeout :param //= 20;
+      ...
+   }
+
+   # The default of 20 will apply here too
+   my $act = Action->new( timeout => $opts{timeout} );
+
+Note that C<$self> is specifically I<not> visible during an initialiser
+expression. This is because the object is not yet fully constructed, so it
+would be dangerous to allow access to it while in this state. However, the
+C<__CLASS__> keyword is available, so initialiser expressions can make use of
+class-based dispatch to invoke class-level methods to help provide values.
+
+This feature should be considered partly B<experimental> and may emit warnings
+to that effect. They can be silenced with
+
+   use Object::Pad ':experimental(init_expr)';
+
+I<Since version 0.76> expressions that are purely compiletime constants
+(either as single scalars or entire lists of constants) are no longer
+considered experimental. They can be used without silencing the warning.
+
+Control flow that attempts to leave a field initialiser expression or block is
+not permitted. This includes any C<return> expression, any C<next/last/redo>
+outside of a loop, with a dynamically-calculated label expression, or with a
+label that it doesn't appear in. C<goto> statements are also currently
+forbidden, though known-safe ones may be permitted in future.
+
+Loop control expressions that are known at compiletime to affect a loop that
+they appear within are permitted.
+
+   field $x { foreach(@list) { next; } }       # this is fine
+
+   field $x { LOOP: while(1) { last LOOP; } }  # this is fine too
+
+=head2 has
+
+   has $var;
+   has @var;
+   has %var;
+
+   has $var = EXPR;
+
+   has $var { BLOCK }
+
+An older version of the L</field> keyword.
+
+This generally behaves like C<field>, except that inline expressions are
+evaluated immediately, once, during class declaration time. These are I<not>
+stored to be evaluated for each constructor.
+
+Because of the one-shot immediate nature of these initialisation expressions
+(and a bunch of other reasons), the C<has> keyword is now discouraged for use.
+Use the C<field> keyword instead.
+
+If you need to evaluate an expression exactly once during the class
+declaration and assign its now-constant value to every instace, store it in a
+regular C<my> variable instead:
+
+   my $default_var = EXPR;
+   field $var = $default_var;
+
 =head2 method
 
    method NAME {
@@ -638,7 +727,7 @@ is C<:lvalue>, allowing easy creation of read-write accessors for fields (but
 see also the C<:reader>, C<:writer> and C<:mutator> field attributes).
 
    class Counter {
-      has $count;
+      field $count;
 
       method count :lvalue { $count }
    }
@@ -660,6 +749,16 @@ Marks that this method expects to override another of the same name from a
 superclass. It is an error at compiletime if the superclass does not provide
 such a method.
 
+=head3 :common
+
+I<Since version 0.62.>
+
+Marks that this method is a class-common method, instead of a regular instance
+method. A class-common method may be invoked on class names instead of
+instances. Within the method body there is a lexical C<$class> available,
+rather than C<$self>. Because it is not associated with a particular object
+instance, a class-common method cannot see instance fields.
+
 =head2 method (lexical)
 
    method $var { ... }
@@ -675,7 +774,7 @@ subsequent method code in the same block by using C<< $self->$var(...) >>
 method call syntax.
 
    class WithPrivate {
-      has $var;
+      field $var;
 
       # Lexical methods can still see instance fields as normal
       method $inc_var { $var++; say "Var was incremented"; }
@@ -734,37 +833,165 @@ I<Since version 0.43.>
 Declares an adjust block for this component class. This block of code runs
 within the constructor, after any C<BUILD> blocks and automatic field value
 assignment. It can make any final adjustments to the instance (such as
-initialising fields from calculated values). No additional parameters are
-passed.
+initialising fields from calculated values).
 
 An adjust block is not a subroutine and thus is not permitted to use
-subroutine attributes. Note that an C<ADJUST> block is a named phaser block
-and not a method; it does not use the C<sub> or C<method> keyword.
+subroutine attributes (except see below). Note that an C<ADJUST> block is a
+named phaser block and not a method; it does not use the C<sub> or C<method>
+keyword.
+
+Currently, an C<ADJUST> block receives a reference to the hash containing the
+current constructor arguments, as per L</ADJUSTPARAMS> (see below). This was
+added in version 0.66 but will be removed again as it conflicts with the more
+flexible and generally nicer named-parameter C<ADJUST :params> syntax
+(see below). Such uses should be considered deprecated. A warning will be
+printed to indicate this whenever an C<ADJUST> block uses a signature. This
+warning can be quieted by using C<ADJUSTPARAMS> instead. Additionally, a
+warning may be printed on code that attempts to access the params hashref via
+the C<@_> array.
+
+=head2 ADJUST :params
+
+   ADJUST :params ( :$var1, :$var2, ... ) {
+      ...
+   }
+
+   ADJUST :params ( :$var1, :$var2, ..., %varN ) {
+      ...
+   }
+
+I<Since version 0.70.>
+
+An C<ADJUST> block can marked with a C<:params> attribute, meaning that it
+consumes additional constructor parameters by assigning them into lexical
+variables.
+
+This feature should be considered B<experimental>, and will emit warnings to
+that effect. They can be silenced with
+
+   use Object::Pad ':experimental(adjust_params)';
+
+Before the block itself, a list of lexical variables are introduced, inside
+parentheses. The name of each one is preceeded by a colon, and consumes a
+constructor parameter of the same name. These parameters are considered
+"consumed" for the purposes of a C<:strict(params)> check.
+
+A named parameter may be provided with default expression, which is evaluated
+if no matching named argument is provided to the constructor. As with fields,
+if a named parameter has no defaulting expression it becomes a required
+argument to the constructor; an exception is thrown by the constructor if it
+absent.
+
+For example,
+
+   ADJUST :params ( :$x, :$y = "default", :$z ) { ... }
+
+Note here that C<x> and C<z> are required parameters for the constructor of a
+class containing this block, but C<y> is an optional parameter whose value
+will be filled in by the expression if not provided. Because these parameters
+are named and not positional, there is no ordering constraint; required and
+optional parameters can be freely mixed.
+
+Optional parameters can also use the C<//=> and C<||=> operators to provide a
+default expression. In these cases, the default will be applied if the caller
+did not provide the named argument at all, or if the provided value was not
+defined (for C<//=>) or not true (for C<||=>).
+
+   ADJUST :params ( :$name //= "unnamed" ) { ... }
+
+Like with subroutine signature parameters, every declared named parameter is
+visible to the defaulting expression of all the later ones. This permits
+values to be calculated based on other ones. For example,
+
+   ADJUST :params ( :$thing = undef, :$things = [ $thing ] ) {
+      # Here, @$things is a list of values
+   }
+
+This permits the caller to pass a list of values via an array reference in
+the C<things> parameter, or a single value in C<thing>.
+
+The final element may be a regular hash variable. This requests that all
+remaining named parameters are made available inside it. The code in the block
+should C<delete> from this hash any parameters it wishes to consume, as with
+the earlier case above.
+
+It is I<unspecified> whether named fields or parameters for subclasses yet to
+be processed are visible to hashes of earlier superclasses. In the current
+implementation they are, but code should not rely on this fact.
+
+Note also that there must be a space between the C<:params> attribute and the
+parentheses holding the named parameters. If this space is not present, perl
+will parse the parentheses as if they are the value to the C<:params()>
+attribute, and this will fail to parse as intended. As with other attributes
+and subroutine signatures, this whitespace B<is> significant.
+
+(This notation is borrowed from a plan to add named parameter support to
+perl's subroutine signature syntax).
 
 =head2 ADJUSTPARAMS
 
+I<Since version 0.51.>
+
    ADJUSTPARAMS ( $params ) {    # on perl 5.26 onwards
       ...
    }
 
-   ADJUSTPARAMS {
+   ADJUST {
       my $params = shift;
       ...
    }
 
-I<Since version 0.51.>
-
-Declares an adjust block for this component class that receives the parameters
-hash reference. This block of code runs within the constructor at the same
-time as L</ADJUST> blocks, but receives in addition a reference to the hash
+A variant of an C<ADJUST> block that receives a reference to the hash
 containing the current constructor parameters. This hash will not contain any
 constructor parameters already consumed by L</:param> declarations on any
 fields, but only the leftovers once those are processed.
 
 The code in the block should C<delete> from this hash any parameters it wishes
-to consume. Once all the C<ADJUSTPARAMS> blocks have run, any remaining keys
-in the hash will be considered errors, subject to the L</:strict(params)>
-check.
+to consume. Once all the C<ADJUST> blocks have run, any remaining keys in the
+hash will be considered errors, subject to the L</:strict(params)> check.
+
+=head2 __CLASS__
+
+   my $classname = __CLASS__;
+
+I<Since version 0.72.>
+
+Only valid within the body (or signature) of a C<method>, an C<ADJUST> block,
+or the initialising expression of a C<field>. Yields the class name of the
+instance that the method, block or expression is invoked on.
+
+This is similar to the core perl C<__PACKAGE__> constant, except that it cares
+about the dynamic class of the actual instance, not the static class the code
+belongs to. When invoked by a subclass instance that inherited code from its
+superclass it yields the name of the class of the instance regardless of which
+class defined the code.
+
+For example,
+
+   class BaseClass {
+      ADJUST { say "Constructing an instance of " . __CLASS__; }
+   }
+
+   class DerivedClass :isa(BaseClass) { }
+
+   my $obj = DerivedClass->new;
+
+Will produce the following output
+
+   Constructing an instance of DerivedClass
+
+This is particularly useful in field initialisers for invoking (constant)
+methods on the invoking class to provide default values for fields. This way a
+subclass could provide a different value.
+
+   class Timer {
+      use constant DEFAULT_DURATION => 60;
+      field $duration = __CLASS__->DEFAULT_DURATION;
+   }
+
+   class ThreeMinuteTimer :isa(Timer) {
+      use constant DEFAULT_DURATION => 3 * 60;
+   }
 
 =head2 requires
 
@@ -827,7 +1054,8 @@ the constructor. This is supported here since C<Object::Pad> version 0.19.
 Note however that any methods invoked by the superclass constructor may not
 see the object in a fully consistent state. (This fact is not specific to
 using C<Object::Pad> and would happen in classic Perl OO as well). The field
-initialisers will have been invoked but the C<BUILD> blocks will not.
+initialisers will have been invoked but the C<BUILD> and C<ADJUST> blocks will
+not.
 
 For example; in the following
 
@@ -841,8 +1069,8 @@ For example; in the following
    }
 
    class DerivedClass :isa(ClassicPerlBaseClass) {
-      has $_value = "B";
-      BUILD {
+      field $_value = "B";
+      ADJUST {
          $_value = "C";
       }
       method get_value { return $_value }
@@ -852,10 +1080,10 @@ For example; in the following
    say "Value seen by user is ", $obj->get_value;
 
 Until the C<ClassicPerlBaseClass::new> superconstructor has returned the
-C<BUILD> block will not have been invoked. The C<$_value> field will still
+C<ADJUST> block will not have been invoked. The C<$_value> field will still
 exist, but its value will be C<B> during the superconstructor. After the
-superconstructor, the C<BUILD> blocks are invoked before the completed object
-is returned to the user. The result will therefore be:
+superconstructor, the C<BUILD> and C<ADJUST> blocks are invoked before the
+completed object is returned to the user. The result will therefore be:
 
    Value seen by superconstructor is B
    Value seen by user is C
@@ -910,7 +1138,7 @@ be implied by the C<class> keyword.
 
    # other use statements
 
-   # has, methods, etc.. can go here
+   # field, methods, etc.. can go here
 
 =head2 Field Names
 
@@ -919,8 +1147,8 @@ lowercase, name components separated by underscores. For tiny examples such as
 "dumb record" structures this may be sufficient.
 
    class Tag {
-      has $name  :mutator;
-      has $value :mutator;
+      field $name  :mutator;
+      field $value :mutator;
    }
 
 In larger examples with lots of non-trivial method bodies, it can get
@@ -930,7 +1158,7 @@ suggested to prefix the field names with a leading underscore, to make them
 more visually distinct.
 
    class Spudger {
-      has $_grapefruit;
+      field $_grapefruit;
 
       ...
 
@@ -949,29 +1177,95 @@ sub import
    $class->import_into( $caller, @_ );
 }
 
+sub _import_experimental
+{
+   shift;
+   my ( $syms, @experiments ) = @_;
+
+   my %enabled;
+
+   my $i = 0;
+   while( $i < @$syms ) {
+      my $sym = $syms->[$i];
+
+      if( $sym eq ":experimental" ) {
+         $enabled{$_}++ for @experiments;
+      }
+      elsif( $sym =~ m/^:experimental\((.*)\)$/ ) {
+         my $tags = $1 =~ s/^\s+|\s+$//gr; # trim
+         $enabled{$_}++ for split m/\s+/, $tags;
+      }
+      else {
+         $i++;
+         next;
+      }
+
+      splice @$syms, $i, 1, ();
+   }
+
+   foreach ( @experiments ) {
+      $^H{"Object::Pad/experimental($_)"}++ if delete $enabled{$_};
+   }
+
+   croak "Unrecognised :experimental features @{[ keys %enabled ]}" if keys %enabled;
+}
+
+sub _import_configuration
+{
+   shift;
+   my ( $syms ) = @_;
+
+   # Undocumented options, purely to support Feature::Compat::Class adjusting
+   # the behaviour to closer match core's  use feature 'class'
+
+   my $i = 0;
+   while( $i < @$syms ) {
+      my $sym = $syms->[$i];
+
+      if( $sym =~ m/^:config\((.*)\)$/ ) {
+         foreach my $opt ( split m/\s+/, $1 =~ s/^\s+|\s+$//gr ) {
+            if( $opt =~ m/^(only_class_attrs|only_field_attrs)=(.*)$/ ) {
+               # Store an entire sub-hash inside the hints hash. This won't
+               # survive squashing into a COP for runtime but we only need it
+               # during compile so that's OK
+               my ( $name, $attrs ) = ( $1, $2 );
+               $^H{"Object::Pad/configure($name)"} = { map { $_ => 1 } split m/,/, $attrs };
+            }
+            else {
+               $^H{"Object::Pad/configure($opt)"}++
+            }
+         }
+      }
+      else {
+         $i++;
+         next;
+      }
+
+      splice @$syms, $i, 1, ();
+   }
+}
+
 sub import_into
 {
    my $class = shift;
-   my ( $caller, @syms ) = @_;
+   my $caller = shift;
 
-   @syms or @syms = qw( class role method has requires );
+   $class->_import_experimental( \@_, qw( init_expr mop custom_field_attr adjust_params ) );
 
-   my %syms = map { $_ => 1 } @syms;
-   delete $syms{$_} and $^H{"Object::Pad/$_"}++ for qw( class role method has requires );
+   $class->_import_configuration( \@_ );
+
+   my %syms = map { $_ => 1 } @_;
+
+   # Default imports
+   unless( %syms ) {
+      $syms{$_}++ for qw( class role method field has requires BUILD ADJUST );
+   }
+
+   delete $syms{$_} and $^H{"Object::Pad/$_"}++ for qw( class role method field has requires BUILD ADJUST );
 
    croak "Unrecognised import symbols @{[ keys %syms ]}" if keys %syms;
 }
 
-sub begin_class
-{
-   my $class = shift;
-   my ( $name, %args ) = @_;
-
-   Carp::carp "Object::Pad->begin_class is deprecated; use Object::Pad::MOP::Class->begin_class instead";
-
-   Object::Pad::MOP::Class->begin_class( $name, %args );
-}
-
 # The universal base-class methods
 
 sub Object::Pad::UNIVERSAL::BUILDARGS
@@ -984,8 +1278,7 @@ sub Object::Pad::UNIVERSAL::BUILDARGS
 sub Object::Pad::MOP::SlotAttr::register
 {
    shift; # $class
-   carp "Object::Pad::MOP::SlotAttr->register is now deprecated; use Object::Pad::MOP::FieldAttr->register instead";
-   return Object::Pad::MOP::FieldAttr->register( @_ );
+   croak "Object::Pad::MOP::SlotAttr->register is now removed; use Object::Pad::MOP::FieldAttr->register instead";
 }
 
 =head1 WITH OTHER MODULES
@@ -999,7 +1292,7 @@ on object instance fields:
    use Syntax::Keyword::Dynamically;
 
    class Container {
-      has $value = 1;
+      field $value = 1;
 
       method example {
          dynamically $value = 2;
@@ -1032,6 +1325,25 @@ These three modules combine; there is additionally a cross-module test to
 ensure that object instance fields can be C<dynamically> set during a
 suspended C<async method>.
 
+=head2 Devel::MAT
+
+When using L<Devel::MAT> to help analyse or debug memory issues with programs
+that use C<Object::Pad>, you will likely want to additionally install the
+module L<Devel::MAT::Tool::Object::Pad>. This will provide new commands and
+extend existing ones to better assist with analysing details related to
+C<Object::Pad> classes and instances of them.
+
+   pmat> fields 0x55d7c173d4b8
+   The field AV ARRAY(3)=NativeClass at 0x55d7c173d4b8
+   Ix Field   Value
+   0  $sfield SCALAR(UV) at 0x55d7c173d938 = 123
+   ...
+
+   pmat> identify 0x55d7c17606d8
+   REF() at 0x55d7c17606d8 is:
+   └─the %hfield field of ARRAY(3)=NativeClass at 0x55d7c173d4b8, which is:
+   ...
+
 =head1 DESIGN TODOs
 
 The following points are details about the design of pad field-based object
@@ -1041,7 +1353,7 @@ systems in general:
 
 =item *
 
-Is multiple inheritence actually required, if role composition is implemented
+Is multiple inheritance actually required, if role composition is implemented
 including giving roles the ability to use private fields?
 
 =item *
@@ -1064,7 +1376,7 @@ These points are more about this particular module's implementation:
 
 =item *
 
-Consider multiple inheritence of subclassing, if that is still considered
+Consider multiple inheritance of subclassing, if that is still considered
 useful after adding roles.
 
 =item *
@@ -1085,7 +1397,7 @@ L<Syntax::Keyword::Dynamically> instead:
 
    use Syntax::Keyword::Dynamically;
 
-   has $loglevel;
+   field $loglevel;
 
    method quietly {
       dynamically $loglevel = LOG_ERROR;
