@@ -3,7 +3,7 @@
 #
 #  (C) Paul Evans, 2020 -- leonerd@leonerd.org.uk
 
-package Object::Pad::MOP::Method 0.61;
+package Object::Pad::MOP::Method 0.78;
 
 use v5.14;
 use warnings;
@@ -20,8 +20,10 @@ C<Object::Pad::MOP::Method> - meta-object representation of a method of a C<Obje
 Instances of this class represent a method of a class implemented by
 L<Object::Pad>. Accessors provide information about the method.
 
-This API should be considered experimental even within the overall context in
-which C<Object::Pad> is expermental.
+This API should be considered B<experimental>, and will emit warnings to that
+effect. They can be silenced with
+
+   use Object::Pad qw( :experimental(mop) );
 
 =cut
 
@@ -38,6 +40,15 @@ Returns the name of the method, as a plain string.
 Returns the L<Object::Pad::MOP::Class> instance representing the class of
 which this method is a member.
 
+=head2 is_common
+
+   $bool = $metamethod->is_common
+
+I<Since version 0.62.>
+
+Returns true if the method is a class-common method, or false for a regular
+instance method.
+
 =cut
 
 =head1 AUTHOR
