@@ -17,6 +17,14 @@
 #include "optree-additions.c.inc"
 #include "make_argcheck_ops.c.inc"
 #include "newOP_CUSTOM.c.inc"
+#include "OP_HELEMEXISTSOR.c.inc"
+
+#if HAVE_PERL_VERSION(5,36,0)
+#  define HAVE_OP_WEAKEN
+#endif
+
+#define need_PLparser()  ObjectPad__need_PLparser(aTHX)
+void ObjectPad__need_PLparser(pTHX); /* in Object/Pad.xs */
 
 FieldMeta *ObjectPad_mop_create_field(pTHX_ SV *fieldname, ClassMeta *classmeta)
 {
@@ -25,14 +33,11 @@ FieldMeta *ObjectPad_mop_create_field(pTHX_ SV *fieldname, ClassMeta *classmeta)
 
   assert(classmeta->next_fieldix > -1);
 
-  fieldmeta->name = SvREFCNT_inc(fieldname);
-  fieldmeta->class = classmeta;
-  fieldmeta->fieldix = classmeta->next_fieldix;
-  fieldmeta->defaultsv = NULL;
-  fieldmeta->defaultexpr = NULL;
-  fieldmeta->paramname = NULL;
-
-  fieldmeta->hooks = NULL;
+  *fieldmeta = (FieldMeta){
+    .name    = SvREFCNT_inc(fieldname),
+    .class   = classmeta,
+    .fieldix = classmeta->next_fieldix,
+  };
 
   return fieldmeta;
 }
@@ -55,41 +60,79 @@ static void S_mop_field_set_param(pTHX_ FieldMeta *fieldmeta, SV *paramname)
   if(!classmeta->parammap)
     classmeta->parammap = newHV();
 
-  HV *parammap = classmeta->parammap;
-
-  HE *he;
-  if((he = hv_fetch_ent(parammap, paramname, 0, 0))) {
-    ParamMeta *colliding_parammeta = (ParamMeta *)HeVAL(he);
-    if(colliding_parammeta->field->class != classmeta)
-      croak("Already have a named constructor parameter called '%" SVf "' inherited from %" SVf,
-        SVfARG(paramname), SVfARG(colliding_parammeta->field->class->name));
-    else
-      croak("Already have a named constructor parameter called '%" SVf "'", SVfARG(paramname));
-  }
+  check_colliding_param(classmeta, paramname);
 
   ParamMeta *parammeta;
   Newx(parammeta, 1, struct ParamMeta);
 
-  parammeta->name = SvREFCNT_inc(paramname);
-  parammeta->field = fieldmeta;
-  parammeta->fieldix = fieldmeta->fieldix;
+  *parammeta = (struct ParamMeta){
+    .name  = SvREFCNT_inc(paramname),
+    .class = classmeta,
+    .type  = PARAM_FIELD,
+    .field.fieldmeta = fieldmeta,
+    .field.fieldix   = fieldmeta->fieldix,
+  };
 
   fieldmeta->paramname = SvREFCNT_inc(paramname);
 
-  hv_store_ent(parammap, paramname, (SV *)parammeta, 0);
+  hv_store_ent(classmeta->parammap, paramname, (SV *)parammeta, 0);
 }
 
 SV *ObjectPad_mop_field_get_default_sv(pTHX_ FieldMeta *fieldmeta)
 {
-  return fieldmeta->defaultsv;
+  if(!fieldmeta->defaultexpr)
+    return NULL;
+
+  OP *o = fieldmeta->defaultexpr;
+
+  switch(mop_field_get_sigil(fieldmeta)) {
+    case '$':
+      break;
+
+    case '@':
+      if(o->op_type != OP_RV2AV)
+        return NULL;
+      o = cUNOPo->op_first;
+      break;
+
+    case '%':
+      if(o->op_type != OP_RV2HV)
+        return NULL;
+      o = cUNOPo->op_first;
+      break;
+  }
+
+  if(o->op_type != OP_CUSTOM || o->op_ppaddr != PL_ppaddr[OP_CONST])
+    return NULL;
+
+  return cSVOPo_sv;
 }
 
 void ObjectPad_mop_field_set_default_sv(pTHX_ FieldMeta *fieldmeta, SV *sv)
 {
-  if(fieldmeta->defaultsv)
-    SvREFCNT_dec(fieldmeta->defaultsv);
+  if(fieldmeta->defaultexpr)
+    op_free(fieldmeta->defaultexpr);
 
-  fieldmeta->defaultsv = sv;
+  /* An OP_CONST whose op_type is OP_CUSTOM. This way we avoid the opchecker
+   * and finalizer doing bad things to our defaultsv SV by setting it
+   * SvREADONLY_on() */
+  OP *valueop = newSVOP_CUSTOM(PL_ppaddr[OP_CONST], 0, sv);
+
+  switch(mop_field_get_sigil(fieldmeta)) {
+    case '$':
+      fieldmeta->defaultexpr = valueop;
+      break;
+
+    case '@':
+      assert(SvROK(sv) && SvTYPE(SvRV(sv)) == SVt_PVAV);
+      fieldmeta->defaultexpr = newUNOP(OP_RV2AV, 0, valueop);
+      break;
+
+    case '%':
+      assert(SvROK(sv) && SvTYPE(SvRV(sv)) == SVt_PVHV);
+      fieldmeta->defaultexpr = newUNOP(OP_RV2HV, 0, valueop);
+      break;
+  }
 }
 
 typedef struct FieldAttributeRegistration FieldAttributeRegistration;
@@ -111,9 +154,11 @@ static void register_field_attribute(const char *name, const struct FieldHookFun
   FieldAttributeRegistration *reg;
   Newx(reg, 1, struct FieldAttributeRegistration);
 
-  reg->name     = name;
-  reg->funcs    = funcs;
-  reg->funcdata = funcdata;
+  *reg = (struct FieldAttributeRegistration){
+    .name     = name,
+    .funcs    = funcs,
+    .funcdata = funcdata,
+  };
 
   if(funcs->permit_hintkey)
     reg->permit_hintkeylen = strlen(funcs->permit_hintkey);
@@ -140,37 +185,40 @@ void ObjectPad_mop_field_apply_attribute(pTHX_ FieldMeta *fieldmeta, const char
        (!hints || !hv_fetch(hints, reg->funcs->permit_hintkey, reg->permit_hintkeylen, 0)))
       continue;
 
-    if((reg->funcs->flags & OBJECTPAD_FLAG_ATTR_NO_VALUE) && value)
-      croak("Attribute :%s does not permit a value", name);
-    if((reg->funcs->flags & OBJECTPAD_FLAG_ATTR_MUST_VALUE) && !value)
-      croak("Attribute :%s requires a value", name);
-
-    SV *hookdata = value;
-
-    if(reg->funcs->apply) {
-      if(!(*reg->funcs->apply)(aTHX_ fieldmeta, value, &hookdata, reg->funcdata))
-        return;
-    }
-
-    if(hookdata && hookdata == value)
-      SvREFCNT_inc(hookdata);
-
-    if(!fieldmeta->hooks)
-      fieldmeta->hooks = newAV();
-
-    struct FieldHook *hook;
-    Newx(hook, 1, struct FieldHook);
-
-    hook->funcs = reg->funcs;
-    hook->hookdata = hookdata;
-    hook->funcdata = reg->funcdata;
-
-    av_push(fieldmeta->hooks, (SV *)hook);
-
-    return;
+    break;
   }
 
-  croak("Unrecognised field attribute :%s", name);
+  if(!reg)
+    croak("Unrecognised field attribute :%s", name);
+
+  if((reg->funcs->flags & OBJECTPAD_FLAG_ATTR_NO_VALUE) && value)
+    croak("Attribute :%s does not permit a value", name);
+  if((reg->funcs->flags & OBJECTPAD_FLAG_ATTR_MUST_VALUE) && !value)
+    croak("Attribute :%s requires a value", name);
+
+  SV *attrdata = value;
+
+  if(reg->funcs->apply) {
+    if(!(*reg->funcs->apply)(aTHX_ fieldmeta, value, &attrdata, reg->funcdata))
+      return;
+  }
+
+  if(attrdata && attrdata == value)
+    SvREFCNT_inc(attrdata);
+
+  if(!fieldmeta->hooks)
+    fieldmeta->hooks = newAV();
+
+  struct FieldHook *hook;
+  Newx(hook, 1, struct FieldHook);
+
+  *hook = (struct FieldHook){
+    .funcs    = reg->funcs,
+    .attrdata = attrdata,
+    .funcdata = reg->funcdata,
+  };
+
+  av_push(fieldmeta->hooks, (SV *)hook);
 }
 
 struct FieldHook *ObjectPad_mop_field_get_attribute(pTHX_ FieldMeta *fieldmeta, const char *name)
@@ -210,9 +258,147 @@ struct FieldHook *ObjectPad_mop_field_get_attribute(pTHX_ FieldMeta *fieldmeta,
   return NULL;
 }
 
+AV *ObjectPad_mop_field_get_attribute_values(pTHX_ FieldMeta *fieldmeta, const char *name)
+{
+  COPHH *cophh = CopHINTHASH_get(PL_curcop);
+
+  /* First, work out what hookfuncs the name maps to */
+
+  FieldAttributeRegistration *reg;
+  for(reg = fieldattrs; reg; reg = reg->next) {
+    if(!strEQ(name, reg->name))
+      continue;
+
+    if(reg->funcs->permit_hintkey &&
+        !cophh_fetch_pvn(cophh, reg->funcs->permit_hintkey, reg->permit_hintkeylen, 0, 0))
+      continue;
+
+    break;
+  }
+
+  if(!reg)
+    return NULL;
+
+  /* Now lets see if fieldmeta has one */
+
+  if(!fieldmeta->hooks)
+    return NULL;
+
+  AV *ret = NULL;
+
+  U32 hooki;
+  for(hooki = 0; hooki < av_count(fieldmeta->hooks); hooki++) {
+    struct FieldHook *hook = (struct FieldHook *)AvARRAY(fieldmeta->hooks)[hooki];
+
+    if(hook->funcs != reg->funcs)
+      continue;
+
+    if(!ret)
+      ret = newAV();
+
+    av_push(ret, newSVsv(hook->attrdata));
+  }
+
+  return ret;
+}
+
+#define gen_field_init_op(fieldmeta)  S_gen_field_init_op(aTHX_ fieldmeta)
+static OP *S_gen_field_init_op(pTHX_ FieldMeta *fieldmeta)
+{
+  ClassMeta *classmeta = fieldmeta->class;
+
+  char sigil = SvPV_nolen(fieldmeta->name)[0];
+  OP *op = NULL;
+
+  switch(sigil) {
+    case '$':
+    {
+      OP *valueop = NULL;
+
+      if(fieldmeta->defaultexpr) {
+        valueop = fieldmeta->defaultexpr;
+      }
+
+      if(fieldmeta->paramname) {
+        SV *paramname = fieldmeta->paramname;
+
+        if(!valueop)
+          valueop = newop_croak_from_constructor(
+            newSVpvf("Required parameter '%" SVf "' is missing for %" SVf " constructor",
+              SVfARG(paramname), SVfARG(classmeta->name)));
+
+        OP *helemop =
+          newBINOP(OP_HELEM, 0,
+            newPADxVOP(OP_PADHV, OPf_REF, PADIX_PARAMS),
+            newSVOP(OP_CONST, 0, SvREFCNT_inc(paramname)));
+
+        if(fieldmeta->def_if_undef)
+          /* delete $params{$paramname} // valueop */
+          valueop = newLOGOP(OP_DOR, 0, newUNOP(OP_DELETE, 0, helemop), valueop);
+        else if(fieldmeta->def_if_false)
+          /* delete $params{$paramname} || valueop */
+          valueop = newLOGOP(OP_OR, 0, newUNOP(OP_DELETE, 0, helemop), valueop);
+        else
+          /* Equivalent of
+           *   exists $params{$paramname} ? delete $params{$paramname} : valueop; */
+          valueop = newHELEMEXISTSOROP(OPpHELEMEXISTSOR_DELETE << 8, helemop, valueop);
+      }
+
+      if(valueop)
+        op = newBINOP(OP_SASSIGN, 0,
+          valueop,
+          /* $fields[$idx] */
+          newAELEMOP(OPf_MOD,
+            newPADxVOP(OP_PADAV, OPf_MOD|OPf_REF, PADIX_SLOTS),
+            fieldmeta->fieldix));
+      break;
+    }
+    case '@':
+    case '%':
+    {
+      OP *valueop = NULL;
+      U16 coerceop = (sigil == '%') ? OP_RV2HV : OP_RV2AV;
+
+      if(fieldmeta->defaultexpr) {
+        valueop = fieldmeta->defaultexpr;
+      }
+
+      if(valueop) {
+        /* $fields[$idx]->@* or ->%* */
+        OP *lhs = force_list_keeping_pushmark(newUNOP(coerceop, OPf_MOD|OPf_REF,
+                    newAELEMOP(0,
+                      newPADxVOP(OP_PADAV, OPf_MOD|OPf_REF, PADIX_SLOTS),
+                      fieldmeta->fieldix)));
+
+        op = newBINOP(OP_AASSIGN, 0,
+            force_list_keeping_pushmark(valueop),
+            lhs);
+      }
+      break;
+    }
+
+    default:
+      croak("ARGH: not sure how to handle a field sigil %c\n", sigil);
+  }
+
+  return op;
+}
+
 void ObjectPad_mop_field_seal(pTHX_ FieldMeta *fieldmeta)
 {
   MOP_FIELD_RUN_HOOKS_NOARGS(fieldmeta, seal);
+
+  need_PLparser();
+
+  ClassMeta *classmeta = fieldmeta->class;
+
+  OP *lines = classmeta->initfields_lines;
+
+  /* TODO: grab a COP at the initexpr time */
+  lines = op_append_elem(OP_LINESEQ, lines, newSTATEOP(0, NULL, NULL));
+  lines = op_append_elem(OP_LINESEQ, lines, gen_field_init_op(fieldmeta));
+
+  classmeta->initfields_lines = lines;
 }
 
 /*******************
@@ -221,11 +407,12 @@ void ObjectPad_mop_field_seal(pTHX_ FieldMeta *fieldmeta)
 
 /* :weak */
 
-static void fieldhook_weak_post_construct(pTHX_ FieldMeta *fieldmeta, SV *_hookdata, void *_funcdata, SV *field)
+static void fieldhook_weak_post_construct(pTHX_ FieldMeta *fieldmeta, SV *_attrdata, void *_funcdata, SV *field)
 {
   sv_rvweaken(field);
 }
 
+#ifndef HAVE_OP_WEAKEN
 static XOP xop_weaken;
 static OP *pp_weaken(pTHX)
 {
@@ -233,14 +420,20 @@ static OP *pp_weaken(pTHX)
   sv_rvweaken(POPs);
   return NORMAL;
 }
+#endif
 
-static void fieldhook_weak_gen_accessor(pTHX_ FieldMeta *fieldmeta, SV *hookdata, void *_funcdata, enum AccessorType type, struct AccessorGenerationCtx *ctx)
+static void fieldhook_weak_gen_accessor(pTHX_ FieldMeta *fieldmeta, SV *attrdata, void *_funcdata, enum AccessorType type, struct AccessorGenerationCtx *ctx)
 {
   if(type != ACCESSOR_WRITER)
     return;
 
   ctx->post_bodyops = op_append_list(OP_LINESEQ, ctx->post_bodyops,
-    newUNOP_CUSTOM(&pp_weaken, 0, newPADxVOP(OP_PADSV, 0, ctx->padix)));
+#ifdef HAVE_OP_WEAKEN
+    newUNOP(OP_WEAKEN, 0,
+#else
+    newUNOP_CUSTOM(&pp_weaken, 0,
+#endif
+      newPADxVOP(OP_PADSV, 0, ctx->padix)));
 }
 
 static struct FieldHookFuncs fieldhooks_weak = {
@@ -251,7 +444,7 @@ static struct FieldHookFuncs fieldhooks_weak = {
 
 /* :param */
 
-static bool fieldhook_param_apply(pTHX_ FieldMeta *fieldmeta, SV *value, SV **hookdata_ptr, void *_funcdata)
+static bool fieldhook_param_apply(pTHX_ FieldMeta *fieldmeta, SV *value, SV **attrdata_ptr, void *_funcdata)
 {
   if(SvPVX(fieldmeta->name)[0] != '$')
     croak("Can only add a named constructor parameter for scalar fields");
@@ -274,7 +467,7 @@ static bool fieldhook_param_apply(pTHX_ FieldMeta *fieldmeta, SV *value, SV **ho
 
   mop_field_set_param(fieldmeta, namesv);
 
-  *hookdata_ptr = namesv;
+  *attrdata_ptr = namesv;
   return TRUE;
 }
 
@@ -315,6 +508,14 @@ static void S_generate_field_accessor_method(pTHX_ FieldMeta *fieldmeta, SV *mna
 
   SV *mname_fq = newSVpvf("%" SVf "::%" SVf, classmeta->name, mname);
 
+  if(PL_curstash != classmeta->stash) {
+    /* RT141599 */
+    SAVESPTR(PL_curstash);
+    PL_curstash = classmeta->stash;
+  }
+
+  need_PLparser();
+
   I32 floor_ix = start_subparse(FALSE, 0);
   SAVEFREESV(PL_compcv);
 
@@ -379,26 +580,26 @@ static void S_generate_field_accessor_method(pTHX_ FieldMeta *fieldmeta, SV *mna
   SvREFCNT_inc(PL_compcv);
   ops = block_end(save_ix, ops);
 
-  CV *cv = newATTRSUB(floor_ix, newSVOP(OP_CONST, 0, mname_fq), NULL, NULL, ops);
+  CV *cv = newATTRSUB(floor_ix, NULL, NULL, NULL, ops);
   CvMETHOD_on(cv);
 
-  mop_class_add_method(classmeta, mname);
+  mop_class_add_method_cv(classmeta, mname, cv);
 
   LEAVE;
 }
 
-static bool fieldhook_reader_apply(pTHX_ FieldMeta *fieldmeta, SV *value, SV **hookdata_ptr, void *_funcdata)
+static bool fieldhook_reader_apply(pTHX_ FieldMeta *fieldmeta, SV *value, SV **attrdata_ptr, void *_funcdata)
 {
-  *hookdata_ptr = make_accessor_mnamesv(aTHX_ fieldmeta, value, "%s");
+  *attrdata_ptr = make_accessor_mnamesv(aTHX_ fieldmeta, value, "%s");
   return TRUE;
 }
 
-static void fieldhook_reader_seal(pTHX_ FieldMeta *fieldmeta, SV *hookdata, void *_funcdata)
+static void fieldhook_reader_seal(pTHX_ FieldMeta *fieldmeta, SV *attrdata, void *_funcdata)
 {
-  S_generate_field_accessor_method(aTHX_ fieldmeta, hookdata, ACCESSOR_READER);
+  S_generate_field_accessor_method(aTHX_ fieldmeta, attrdata, ACCESSOR_READER);
 }
 
-static void fieldhook_gen_reader_ops(pTHX_ FieldMeta *fieldmeta, SV *hookdata, void *_funcdata, enum AccessorType type, struct AccessorGenerationCtx *ctx)
+static void fieldhook_gen_reader_ops(pTHX_ FieldMeta *fieldmeta, SV *attrdata, void *_funcdata, enum AccessorType type, struct AccessorGenerationCtx *ctx)
 {
   if(type != ACCESSOR_READER)
     return;
@@ -425,18 +626,18 @@ static struct FieldHookFuncs fieldhooks_reader = {
 
 /* :writer */
 
-static bool fieldhook_writer_apply(pTHX_ FieldMeta *fieldmeta, SV *value, SV **hookdata_ptr, void *_funcdata)
+static bool fieldhook_writer_apply(pTHX_ FieldMeta *fieldmeta, SV *value, SV **attrdata_ptr, void *_funcdata)
 {
-  *hookdata_ptr = make_accessor_mnamesv(aTHX_ fieldmeta, value, "set_%s");
+  *attrdata_ptr = make_accessor_mnamesv(aTHX_ fieldmeta, value, "set_%s");
   return TRUE;
 }
 
-static void fieldhook_writer_seal(pTHX_ FieldMeta *fieldmeta, SV *hookdata, void *_funcdata)
+static void fieldhook_writer_seal(pTHX_ FieldMeta *fieldmeta, SV *attrdata, void *_funcdata)
 {
-  S_generate_field_accessor_method(aTHX_ fieldmeta, hookdata, ACCESSOR_WRITER);
+  S_generate_field_accessor_method(aTHX_ fieldmeta, attrdata, ACCESSOR_WRITER);
 }
 
-static void fieldhook_gen_writer_ops(pTHX_ FieldMeta *fieldmeta, SV *hookdata, void *_funcdata, enum AccessorType type, struct AccessorGenerationCtx *ctx)
+static void fieldhook_gen_writer_ops(pTHX_ FieldMeta *fieldmeta, SV *attrdata, void *_funcdata, enum AccessorType type, struct AccessorGenerationCtx *ctx)
 {
   if(type != ACCESSOR_WRITER)
     return;
@@ -475,22 +676,22 @@ static struct FieldHookFuncs fieldhooks_writer = {
 
 /* :mutator */
 
-static bool fieldhook_mutator_apply(pTHX_ FieldMeta *fieldmeta, SV *value, SV **hookdata_ptr, void *_funcdata)
+static bool fieldhook_mutator_apply(pTHX_ FieldMeta *fieldmeta, SV *value, SV **attrdata_ptr, void *_funcdata)
 {
   if(SvPVX(fieldmeta->name)[0] != '$')
     /* TODO: A reader for an array or hash field should also be fine */
     croak("Can only generate accessors for scalar fields");
 
-  *hookdata_ptr = make_accessor_mnamesv(aTHX_ fieldmeta, value, "%s");
+  *attrdata_ptr = make_accessor_mnamesv(aTHX_ fieldmeta, value, "%s");
   return TRUE;
 }
 
-static void fieldhook_mutator_seal(pTHX_ FieldMeta *fieldmeta, SV *hookdata, void *_funcdata)
+static void fieldhook_mutator_seal(pTHX_ FieldMeta *fieldmeta, SV *attrdata, void *_funcdata)
 {
-  S_generate_field_accessor_method(aTHX_ fieldmeta, hookdata, ACCESSOR_LVALUE_MUTATOR);
+  S_generate_field_accessor_method(aTHX_ fieldmeta, attrdata, ACCESSOR_LVALUE_MUTATOR);
 }
 
-static void fieldhook_gen_mutator_ops(pTHX_ FieldMeta *fieldmeta, SV *hookdata, void *_funcdata, enum AccessorType type, struct AccessorGenerationCtx *ctx)
+static void fieldhook_gen_mutator_ops(pTHX_ FieldMeta *fieldmeta, SV *attrdata, void *_funcdata, enum AccessorType type, struct AccessorGenerationCtx *ctx)
 {
   if(type != ACCESSOR_LVALUE_MUTATOR)
     return;
@@ -511,12 +712,12 @@ static struct FieldHookFuncs fieldhooks_mutator = {
 
 /* :accessor */
 
-static void fieldhook_accessor_seal(pTHX_ FieldMeta *fieldmeta, SV *hookdata, void *_funcdata)
+static void fieldhook_accessor_seal(pTHX_ FieldMeta *fieldmeta, SV *attrdata, void *_funcdata)
 {
-  S_generate_field_accessor_method(aTHX_ fieldmeta, hookdata, ACCESSOR_COMBINED);
+  S_generate_field_accessor_method(aTHX_ fieldmeta, attrdata, ACCESSOR_COMBINED);
 }
 
-static void fieldhook_gen_accessor_ops(pTHX_ FieldMeta *fieldmeta, SV *hookdata, void *_funcdata, enum AccessorType type, struct AccessorGenerationCtx *ctx)
+static void fieldhook_gen_accessor_ops(pTHX_ FieldMeta *fieldmeta, SV *attrdata, void *_funcdata, enum AccessorType type, struct AccessorGenerationCtx *ctx)
 {
   if(type != ACCESSOR_COMBINED)
     return;
@@ -562,10 +763,12 @@ void ObjectPad_register_field_attribute(pTHX_ const char *name, const struct Fie
 
 void ObjectPad__boot_fields(pTHX)
 {
+#ifndef HAVE_OP_WEAKEN
   XopENTRY_set(&xop_weaken, xop_name, "weaken");
   XopENTRY_set(&xop_weaken, xop_desc, "weaken an RV");
   XopENTRY_set(&xop_weaken, xop_class, OA_UNOP);
   Perl_custom_op_register(aTHX_ &pp_weaken, &xop_weaken);
+#endif
 
   register_field_attribute("weak",     &fieldhooks_weak,     NULL);
   register_field_attribute("param",    &fieldhooks_param,    NULL);
@@ -574,51 +777,3 @@ void ObjectPad__boot_fields(pTHX)
   register_field_attribute("mutator",  &fieldhooks_mutator,  NULL);
   register_field_attribute("accessor", &fieldhooks_accessor, NULL);
 }
-
-
-/* back-compat */
-FieldMeta *ObjectPad_mop_create_slot(pTHX_ SV *fieldname, ClassMeta *classmeta)
-{
-  return ObjectPad_mop_create_field(aTHX_ fieldname, classmeta);
-}
-
-void ObjectPad_mop_slot_seal(pTHX_ FieldMeta *fieldmeta)
-{
-  return ObjectPad_mop_field_seal(aTHX_ fieldmeta);
-}
-
-SV *ObjectPad_mop_slot_get_name(pTHX_ FieldMeta *fieldmeta)
-{
-  return ObjectPad_mop_field_get_name(aTHX_ fieldmeta);
-}
-
-char ObjectPad_mop_slot_get_sigil(pTHX_ FieldMeta *fieldmeta)
-{
-  return ObjectPad_mop_field_get_sigil(aTHX_ fieldmeta);
-}
-
-void ObjectPad_mop_slot_apply_attribute(pTHX_ FieldMeta *fieldmeta, const char *name, SV *value)
-{
-  ObjectPad_mop_field_apply_attribute(aTHX_ fieldmeta, name, value);
-}
-
-struct FieldHook *ObjectPad_mop_slot_get_attribute(pTHX_ FieldMeta *fieldmeta, const char *name)
-{
-  return ObjectPad_mop_field_get_attribute(aTHX_ fieldmeta, name);
-}
-
-SV *ObjectPad_mop_slot_get_default_sv(pTHX_ FieldMeta *fieldmeta)
-{
-  return ObjectPad_mop_field_get_default_sv(aTHX_ fieldmeta);
-}
-
-void ObjectPad_mop_slot_set_default_sv(pTHX_ FieldMeta *fieldmeta, SV *sv)
-{
-  ObjectPad_mop_field_set_default_sv(aTHX_ fieldmeta, sv);
-}
-
-void ObjectPad_register_slot_attribute(pTHX_ const char *name, const struct SlotHookFuncs *funcs, void *funcdata)
-{
-  Perl_warn(aTHX_ "register_slot_attribute() is now deprecated; use register_field_attribute() instead");
-  ObjectPad_register_field_attribute(aTHX_ name, (const struct FieldHookFuncs *)funcs, funcdata);
-}
